import re
import orjson
import tempfile
import subprocess
from pathlib import Path
from http.cookies import SimpleCookie

from bbot.modules.base import BaseModule


class httpx(BaseModule):
    watched_events = ["OPEN_TCP_PORT", "URL_UNVERIFIED", "URL"]
    produced_events = ["URL", "HTTP_RESPONSE"]
    flags = ["active", "safe", "web-basic", "social-enum", "subdomain-enum", "cloud-enum"]
    meta = {
        "description": "Visit webpages. Many other modules rely on httpx",
        "created_date": "2022-07-08",
        "author": "@TheTechromancer",
    }

    options = {
        "threads": 50,
        "in_scope_only": True,
        "version": "1.2.5",
        "max_response_size": 5242880,
        "store_responses": False,
        "probe_all_ips": False,
    }
    options_desc = {
        "threads": "Number of httpx threads to use",
        "in_scope_only": "Only visit web reparents that are in scope.",
        "version": "httpx version",
        "max_response_size": "Max response size in bytes",
        "store_responses": "Save raw HTTP responses to scan folder",
        "probe_all_ips": "Probe all the ips associated with same host",
    }
    deps_ansible = [
        {
            "name": "Download httpx",
            "unarchive": {
                "src": "https://github.com/projectdiscovery/httpx/releases/download/v#{BBOT_MODULES_HTTPX_VERSION}/httpx_#{BBOT_MODULES_HTTPX_VERSION}_#{BBOT_OS}_#{BBOT_CPU_ARCH_GOLANG}.zip",
                "include": "httpx",
                "dest": "#{BBOT_TOOLS}",
                "remote_src": True,
            },
        }
    ]

    scope_distance_modifier = 2
    _shuffle_incoming_queue = False
    _batch_size = 500
    _priority = 2
    # accept Javascript URLs
    accept_url_special = True

    async def setup(self):
        self.threads = self.config.get("threads", 50)
        self.max_response_size = self.config.get("max_response_size", 5242880)
        self.store_responses = self.config.get("store_responses", False)
        self.probe_all_ips = self.config.get("probe_all_ips", False)
        self.httpx_tempdir_regex = re.compile(r"^httpx\d+$")
        return True

    async def filter_event(self, event):
        if "_wildcard" in str(event.host).split("."):
            return False, "event is wildcard"

        if "unresolved" in event.tags:
            return False, "event is unresolved"

        if event.module == self:
            return False, "event is from self"

        if "spider-max" in event.tags:
            return False, "event exceeds spidering limits"

        # scope filtering
        in_scope_only = self.config.get("in_scope_only", True)
        if "httpx-safe" in event.tags:
            return True
        max_scope_distance = 0 if in_scope_only else (self.scan.scope_search_distance + 1)
        if event.scope_distance > max_scope_distance:
            return False, "event is not in scope"
        return True

    def make_url_metadata(self, event):
        has_spider_max = "spider-max" in event.tags
        url_hash = None
        if event.type.startswith("URL"):
            # we NEED the port, otherwise httpx will try HTTPS even for HTTP URLs
            url = event.with_port().geturl()
            if event.parsed_url.path == "/":
                url_hash = hash((event.host, event.port, has_spider_max))
        else:
            url = str(event.data)
            url_hash = hash((event.host, event.port, has_spider_max))
        if url_hash is None:
            url_hash = hash((url, has_spider_max))
        return url, url_hash

    def _incoming_dedup_hash(self, event):
        url, url_hash = self.make_url_metadata(event)
        return url_hash

    async def handle_batch(self, *events):
        stdin = {}

        for event in events:
            url, url_hash = self.make_url_metadata(event)
            stdin[url] = event

        if not stdin:
            return

        command = [
            "httpx",
            "-silent",
            "-json",
            "-include-response",
            "-threads",
            self.threads,
            "-timeout",
            self.scan.httpx_timeout,
            "-retries",
            self.scan.httpx_retries,
            "-header",
            f"User-Agent: {self.scan.useragent}",
            "-response-size-to-read",
            f"{self.max_response_size}",
        ]

        if self.store_responses:
            response_dir = self.scan.home / "httpx"
            self.helpers.mkdir(response_dir)
            command += ["-srd", str(response_dir)]

        dns_resolvers = ",".join(self.helpers.system_resolvers)
        if dns_resolvers:
            command += ["-r", dns_resolvers]

        if self.probe_all_ips:
            command += ["-probe-all-ips"]

        # Add custom HTTP headers
        for hk, hv in self.scan.custom_http_headers.items():
            command += ["-header", f"{hk}: {hv}"]

        # Add custom HTTP cookies as a single header
        if self.scan.custom_http_cookies:
            cookie = SimpleCookie()
            for ck, cv in self.scan.custom_http_cookies.items():
                cookie[ck] = cv

            # Build the cookie header
            cookie_header = f"Cookie: {cookie.output(header='', sep='; ').strip()}"
            command += ["-header", cookie_header]

        proxy = self.scan.http_proxy
        if proxy:
            command += ["-http-proxy", proxy]
        async for line in self.run_process_live(command, text=False, input=list(stdin), stderr=subprocess.DEVNULL):
            try:
                j = await self.helpers.run_in_executor(orjson.loads, line)
            except orjson.JSONDecodeError:
                self.warning(f"httpx failed to decode line: {line}")
                continue

            url = j.get("url", "")
            status_code = int(j.get("status_code", 0))
            if status_code == 0:
                self.debug(f'No HTTP status code for "{url}"')
                continue

            parent_event = stdin.get(j.get("input", ""), None)

            if parent_event is None:
                self.warning(f"Unable to correlate parent event from: {line}")
                continue

            # discard 404s from unverified URLs
            path = j.get("path", "/")
            if parent_event.type == "URL_UNVERIFIED" and status_code in (404,) and path != "/":
                self.debug(f'Discarding 404 from "{url}"')
                continue

            # main URL
            tags = [f"status-{status_code}"]
            httpx_ip = j.get("host", "")
            if httpx_ip:
                tags.append(f"ip-{httpx_ip}")
            # grab title
            title = self.helpers.tagify(j.get("title", ""), maxlen=30)
            if title:
                tags.append(f"http-title-{title}")

            url_context = "{module} visited {event.parent.data} and got status code {event.http_status}"
            if parent_event.type == "OPEN_TCP_PORT":
                url_context += " at {event.data}"

            url_event = self.make_event(
                url,
                "URL",
                parent_event,
                tags=tags,
                context=url_context,
            )
            if url_event:
                if url_event != parent_event:
                    await self.emit_event(url_event)
                # HTTP response
                content_type = j.get("header", {}).get("content_type", "unspecified").split(";")[0]
                content_length = j.get("content_length", 0)
                content_length = self.helpers.bytes_to_human(content_length)
                await self.emit_event(
                    j,
                    "HTTP_RESPONSE",
                    url_event,
                    tags=url_event.tags,
                    context=f"HTTP_RESPONSE was {content_length} with {content_type} content type",
                )

        for tempdir in Path(tempfile.gettempdir()).iterdir():
            if tempdir.is_dir() and self.httpx_tempdir_regex.match(tempdir.name):
                self.helpers.rm_rf(tempdir)

    async def cleanup(self):
        resume_file = self.helpers.current_dir / "resume.cfg"
        resume_file.unlink(missing_ok=True)



import re
import logging
import argparse
from omegaconf import OmegaConf

from bbot.errors import *
from bbot.core.helpers.misc import chain_lists, get_closest_match, get_keys_in_dot_syntax

log = logging.getLogger("bbot.presets.args")


universal_module_options = {
    "batch_size": "The number of events to process in a single batch (only applies to batch modules)",
    "module_threads": "How many event handlers to run in parallel",
    "module_timeout": "Max time in seconds to spend handling each event or batch of events",
}


class BBOTArgs:
    # module config options to exclude from validation
    exclude_from_validation = re.compile(
        r".*modules\.[a-z0-9_]+\.(?:" + "|".join(universal_module_options.keys()) + ")$"
    )

    scan_examples = [
        (
            "Subdomains",
            "Perform a full subdomain enumeration on evilcorp.com",
            "bbot -t evilcorp.com -p subdomain-enum",
        ),
        (
            "Subdomains (passive only)",
            "Perform a passive-only subdomain enumeration on evilcorp.com",
            "bbot -t evilcorp.com -p subdomain-enum -rf passive",
        ),
        (
            "Subdomains + port scan + web screenshots",
            "Port-scan every subdomain, screenshot every webpage, output to current directory",
            "bbot -t evilcorp.com -p subdomain-enum -m portscan gowitness -n my_scan -o .",
        ),
        (
            "Subdomains + basic web scan",
            "A basic web scan includes robots.txt, storage buckets, IIS shortnames, and other non-intrusive web modules",
            "bbot -t evilcorp.com -p subdomain-enum web-basic",
        ),
        (
            "Web spider",
            "Crawl www.evilcorp.com up to a max depth of 2, automatically extracting emails, secrets, etc.",
            "bbot -t www.evilcorp.com -p spider -c web.spider_distance=2 web.spider_depth=2",
        ),
        (
            "Everything everywhere all at once",
            "Subdomains, emails, cloud buckets, port scan, basic web, web screenshots, nuclei",
            "bbot -t evilcorp.com -p kitchen-sink",
        ),
    ]

    usage_examples = [
        (
            "List modules",
            "",
            "bbot -l",
        ),
        (
            "List output modules",
            "",
            "bbot -lo",
        ),
        (
            "List presets",
            "",
            "bbot -lp",
        ),
        (
            "List flags",
            "",
            "bbot -lf",
        ),
        (
            "Show help for a specific module",
            "",
            "bbot -mh <module_name>",
        ),
    ]

    epilog = "EXAMPLES\n"
    for example in (scan_examples, usage_examples):
        for title, description, command in example:
            epilog += f"\n    {title}:\n        {command}\n"

    def __init__(self, preset):
        self.preset = preset
        self._config = None

        self.parser = self.create_parser()
        self._parsed = None

    @property
    def parsed(self):
        if self._parsed is None:
            self._parsed = self.parser.parse_args()
            self.sanitize_args()
        return self._parsed

    def preset_from_args(self):
        # the order here is important
        # first we make the preset
        args_preset = self.preset.__class__(
            *self.parsed.targets,
            whitelist=self.parsed.whitelist,
            blacklist=self.parsed.blacklist,
            name="args_preset",
        )

        # then we load requested preset
        # this is important so we can load custom module directories, pull in custom flags, module config options, etc.
        for preset_arg in self.parsed.preset:
            try:
                args_preset.include_preset(preset_arg)
            except BBOTArgumentError:
                raise
            except Exception as e:
                raise BBOTArgumentError(f'Error parsing preset "{preset_arg}": {e}')

        # then we set verbosity levels (so if the user enables -d they can see debug output)
        if self.parsed.silent:
            args_preset.silent = True
        if self.parsed.verbose:
            args_preset.verbose = True
        if self.parsed.debug:
            args_preset.debug = True

        # modules + flags
        args_preset.exclude_modules.update(set(self.parsed.exclude_modules))
        args_preset.exclude_flags.update(set(self.parsed.exclude_flags))
        args_preset.require_flags.update(set(self.parsed.require_flags))
        args_preset.explicit_scan_modules.update(set(self.parsed.modules))
        args_preset.explicit_output_modules.update(set(self.parsed.output_modules))
        args_preset.flags.update(set(self.parsed.flags))

        # output
        if self.parsed.json:
            args_preset.core.merge_custom({"modules": {"stdout": {"format": "json"}}})
        if self.parsed.brief:
            args_preset.core.merge_custom(
                {"modules": {"stdout": {"event_fields": ["type", "scope_description", "data"]}}}
            )
        if self.parsed.event_types:
            args_preset.core.merge_custom({"modules": {"stdout": {"event_types": self.parsed.event_types}}})
        if self.parsed.exclude_cdn:
            args_preset.explicit_scan_modules.add("portfilter")

        # dependencies
        deps_config = args_preset.core.custom_config.get("deps", {})
        if self.parsed.retry_deps:
            deps_config["behavior"] = "retry_failed"
        elif self.parsed.force_deps:
            deps_config["behavior"] = "force_install"
        elif self.parsed.no_deps:
            deps_config["behavior"] = "disable"
        elif self.parsed.ignore_failed_deps:
            deps_config["behavior"] = "ignore_failed"
        if deps_config:
            args_preset.core.merge_custom({"deps": deps_config})

        # other scan options
        if self.parsed.name is not None:
            args_preset.scan_name = self.parsed.name
        if self.parsed.output_dir is not None:
            args_preset.output_dir = self.parsed.output_dir
        if self.parsed.force:
            args_preset.force_start = self.parsed.force

        if self.parsed.proxy:
            args_preset.core.merge_custom({"web": {"http_proxy": self.parsed.proxy}})

        if self.parsed.custom_headers:
            args_preset.core.merge_custom({"web": {"http_headers": self.parsed.custom_headers}})

        if self.parsed.custom_cookies:
            args_preset.core.merge_custom({"web": {"http_cookies": self.parsed.custom_cookies}})

        if self.parsed.custom_yara_rules:
            args_preset.core.merge_custom(
                {"modules": {"excavate": {"custom_yara_rules": self.parsed.custom_yara_rules}}}
            )

        # Check if both user_agent and user_agent_suffix are set. If so combine them and merge into the config
        if self.parsed.user_agent and self.parsed.user_agent_suffix:
            modified_user_agent = f"{self.parsed.user_agent} {self.parsed.user_agent_suffix}"
            args_preset.core.merge_custom({"web": {"user_agent": modified_user_agent}})

        # If only user_agent_suffix is set, retrieve the existing user_agent from the merged config and append the suffix
        elif self.parsed.user_agent_suffix:
            existing_user_agent = args_preset.core.config.get("web", {}).get("user_agent", "")
            modified_user_agent = f"{existing_user_agent} {self.parsed.user_agent_suffix}"
            args_preset.core.merge_custom({"web": {"user_agent": modified_user_agent}})

        # If only user_agent is set, merge it directly
        elif self.parsed.user_agent:
            args_preset.core.merge_custom({"web": {"user_agent": self.parsed.user_agent}})

        # CLI config options (dot-syntax)
        for config_arg in self.parsed.config:
            try:
                # if that fails, try to parse as key=value syntax
                args_preset.core.merge_custom(OmegaConf.from_cli([config_arg]))
            except Exception as e:
                raise BBOTArgumentError(f'Error parsing command-line config option: "{config_arg}": {e}')

        # strict scope
        if self.parsed.strict_scope:
            args_preset.core.merge_custom({"scope": {"strict": True}})

        return args_preset

    def create_parser(self, *args, **kwargs):
        kwargs.update(
            {
                "description": "Bighuge BLS OSINT Tool",
                "formatter_class": argparse.RawTextHelpFormatter,
                "epilog": self.epilog,
            }
        )
        p = argparse.ArgumentParser(*args, **kwargs)

        target = p.add_argument_group(title="Target")
        target.add_argument(
            "-t", "--targets", nargs="+", default=[], help="Targets to seed the scan", metavar="TARGET"
        )
        target.add_argument(
            "-w",
            "--whitelist",
            nargs="+",
            default=None,
            help="What's considered in-scope (by default it's the same as --targets)",
        )
        target.add_argument("-b", "--blacklist", nargs="+", default=[], help="Don't touch these things")
        target.add_argument(
            "--strict-scope",
            action="store_true",
            help="Don't consider subdomains of target/whitelist to be in-scope",
        )
        presets = p.add_argument_group(title="Presets")
        presets.add_argument(
            "-p",
            "--preset",
            nargs="*",
            help="Enable BBOT preset(s)",
            metavar="PRESET",
            default=[],
        )
        presets.add_argument(
            "-c",
            "--config",
            nargs="*",
            help="Custom config options in key=value format: e.g. 'modules.shodan.api_key=1234'",
            metavar="CONFIG",
            default=[],
        )
        presets.add_argument("-lp", "--list-presets", action="store_true", help="List available presets.")

        modules = p.add_argument_group(title="Modules")
        modules.add_argument(
            "-m",
            "--modules",
            nargs="+",
            default=[],
            help=f"Modules to enable. Choices: {','.join(sorted(self.preset.module_loader.scan_module_choices))}",
            metavar="MODULE",
        )
        modules.add_argument("-l", "--list-modules", action="store_true", help="List available modules.")
        modules.add_argument(
            "-lmo", "--list-module-options", action="store_true", help="Show all module config options"
        )
        modules.add_argument(
            "-em", "--exclude-modules", nargs="+", default=[], help="Exclude these modules.", metavar="MODULE"
        )
        modules.add_argument(
            "-f",
            "--flags",
            nargs="+",
            default=[],
            help=f"Enable modules by flag. Choices: {','.join(sorted(self.preset.module_loader.flag_choices))}",
            metavar="FLAG",
        )
        modules.add_argument("-lf", "--list-flags", action="store_true", help="List available flags.")
        modules.add_argument(
            "-rf",
            "--require-flags",
            nargs="+",
            default=[],
            help="Only enable modules with these flags (e.g. -rf passive)",
            metavar="FLAG",
        )
        modules.add_argument(
            "-ef",
            "--exclude-flags",
            nargs="+",
            default=[],
            help="Disable modules with these flags. (e.g. -ef aggressive)",
            metavar="FLAG",
        )
        modules.add_argument("--allow-deadly", action="store_true", help="Enable the use of highly aggressive modules")

        scan = p.add_argument_group(title="Scan")
        scan.add_argument("-n", "--name", help="Name of scan (default: random)", metavar="SCAN_NAME")
        scan.add_argument("-v", "--verbose", action="store_true", help="Be more verbose")
        scan.add_argument("-d", "--debug", action="store_true", help="Enable debugging")
        scan.add_argument("-s", "--silent", action="store_true", help="Be quiet")
        scan.add_argument(
            "--force",
            action="store_true",
            help="Run scan even in the case of condition violations or failed module setups",
        )
        scan.add_argument("-y", "--yes", action="store_true", help="Skip scan confirmation prompt")
        scan.add_argument(
            "--fast-mode",
            action="store_true",
            help="Scan only the provided targets as fast as possible, with no extra discovery",
        )
        scan.add_argument("--dry-run", action="store_true", help="Abort before executing scan")
        scan.add_argument(
            "--current-preset",
            action="store_true",
            help="Show the current preset in YAML format",
        )
        scan.add_argument(
            "--current-preset-full",
            action="store_true",
            help="Show the current preset in its full form, including defaults",
        )

        scan.add_argument(
            "-mh",
            "--module-help",
            default=None,
            help="Show help for a specific module",
            metavar="MODULE",
        )

        output = p.add_argument_group(title="Output")
        output.add_argument(
            "-o",
            "--output-dir",
            help="Directory to output scan results",
            metavar="DIR",
        )
        output.add_argument(
            "-om",
            "--output-modules",
            nargs="+",
            default=[],
            help=f"Output module(s). Choices: {','.join(sorted(self.preset.module_loader.output_module_choices))}",
            metavar="MODULE",
        )
        output.add_argument("-lo", "--list-output-modules", action="store_true", help="List available output modules")
        output.add_argument("--json", "-j", action="store_true", help="Output scan data in JSON format")
        output.add_argument("--brief", "-br", action="store_true", help="Output only the data itself")
        output.add_argument("--event-types", nargs="+", default=[], help="Choose which event types to display")
        output.add_argument(
            "--exclude-cdn",
            "-ec",
            action="store_true",
            help="Filter out unwanted open ports on CDNs/WAFs (80,443 only)",
        )

        deps = p.add_argument_group(
            title="Module dependencies", description="Control how modules install their dependencies"
        )
        g2 = deps.add_mutually_exclusive_group()
        g2.add_argument("--no-deps", action="store_true", help="Don't install module dependencies")
        g2.add_argument("--force-deps", action="store_true", help="Force install all module dependencies")
        g2.add_argument("--retry-deps", action="store_true", help="Try again to install failed module dependencies")
        g2.add_argument(
            "--ignore-failed-deps", action="store_true", help="Run modules even if they have failed dependencies"
        )
        g2.add_argument("--install-all-deps", action="store_true", help="Install dependencies for all modules")

        misc = p.add_argument_group(title="Misc")
        misc.add_argument("--version", action="store_true", help="show BBOT version and exit")
        misc.add_argument("--proxy", help="Use this proxy for all HTTP requests", metavar="HTTP_PROXY")
        misc.add_argument(
            "-H",
            "--custom-headers",
            nargs="+",
            default=[],
            help="List of custom headers as key value pairs (header=value).",
        )
        misc.add_argument(
            "-C",
            "--custom-cookies",
            nargs="+",
            default=[],
            help="List of custom cookies as key value pairs (cookie=value).",
        )
        misc.add_argument("--custom-yara-rules", "-cy", help="Add custom yara rules to excavate")

        misc.add_argument("--user-agent", "-ua", help="Set the user-agent for all HTTP requests")
        misc.add_argument("--user-agent-suffix", "-uas", help=argparse.SUPPRESS, metavar="SUFFIX", default=None)
        return p

    def sanitize_args(self):
        # silent implies -y
        if self.parsed.silent:
            self.parsed.yes = True
        # chain_lists allows either comma-separated or space-separated lists
        self.parsed.modules = chain_lists(self.parsed.modules)
        self.parsed.exclude_modules = chain_lists(self.parsed.exclude_modules)
        self.parsed.output_modules = chain_lists(self.parsed.output_modules)
        self.parsed.targets = chain_lists(
            self.parsed.targets, try_files=True, msg="Reading targets from file: {filename}"
        )
        if self.parsed.whitelist is not None:
            self.parsed.whitelist = chain_lists(
                self.parsed.whitelist, try_files=True, msg="Reading whitelist from file: {filename}"
            )
        self.parsed.blacklist = chain_lists(
            self.parsed.blacklist, try_files=True, msg="Reading blacklist from file: {filename}"
        )
        self.parsed.flags = chain_lists(self.parsed.flags)
        self.parsed.exclude_flags = chain_lists(self.parsed.exclude_flags)
        self.parsed.require_flags = chain_lists(self.parsed.require_flags)
        self.parsed.event_types = [t.upper() for t in chain_lists(self.parsed.event_types)]

        # Custom Header Parsing / Validation
        custom_headers_dict = {}
        custom_header_example = "Example: --custom-headers foo=bar foo2=bar2"

        for i in self.parsed.custom_headers:
            parts = i.split("=", 1)
            if len(parts) != 2:
                raise ValidationError(f"Custom headers not formatted correctly (missing '='). {custom_header_example}")
            k, v = parts
            if not k or not v:
                raise ValidationError(
                    f"Custom headers not formatted correctly (missing header name or value). {custom_header_example}"
                )
            custom_headers_dict[k] = v
        self.parsed.custom_headers = custom_headers_dict

        # Custom Cookie Parsing / Validation
        custom_cookies_dict = {}
        custom_cookie_example = "Example: --custom-cookies foo=bar foo2=bar2"

        for i in self.parsed.custom_cookies:
            parts = i.split("=", 1)
            if len(parts) != 2:
                raise ValidationError(f"Custom cookies not formatted correctly (missing '='). {custom_cookie_example}")
            k, v = parts
            if not k or not v:
                raise ValidationError(
                    f"Custom cookies not formatted correctly (missing cookie name or value). {custom_cookie_example}"
                )
            custom_cookies_dict[k] = v
        self.parsed.custom_cookies = custom_cookies_dict

        # --fast-mode
        if self.parsed.fast_mode:
            self.parsed.preset += ["fast"]

    def validate(self):
        # validate config options
        sentinel = object()
        all_options = set(get_keys_in_dot_syntax(self.preset.core.default_config))
        for c in self.parsed.config:
            c = c.split("=")[0].strip()
            v = OmegaConf.select(self.preset.core.default_config, c, default=sentinel)
            # if option isn't in the default config
            if v is sentinel:
                # skip if it's excluded from validation
                if self.exclude_from_validation.match(c):
                    continue
                # otherwise, ensure it exists as a module option
                raise ValidationError(get_closest_match(c, all_options, msg="config option"))



import json
import subprocess
from bbot.modules.base import BaseModule


class fingerprintx(BaseModule):
    watched_events = ["OPEN_TCP_PORT"]
    produced_events = ["PROTOCOL"]
    flags = ["active", "safe", "service-enum", "slow"]
    meta = {
        "description": "Fingerprint exposed services like RDP, SSH, MySQL, etc.",
        "created_date": "2023-01-30",
        "author": "@TheTechromancer",
    }
    options = {"version": "1.1.4"}
    options_desc = {"version": "fingerprintx version"}
    _batch_size = 10
    _module_threads = 2
    _priority = 2

    options = {"skip_common_web": True}
    options_desc = {"skip_common_web": "Skip common web ports such as 80, 443, 8080, 8443, etc."}

    deps_ansible = [
        {
            "name": "Download fingerprintx",
            "unarchive": {
                "src": "https://github.com/praetorian-inc/fingerprintx/releases/download/v#{BBOT_MODULES_FINGERPRINTX_VERSION}/fingerprintx_#{BBOT_MODULES_FINGERPRINTX_VERSION}_#{BBOT_OS_PLATFORM}_#{BBOT_CPU_ARCH_GOLANG}.tar.gz",
                "include": "fingerprintx",
                "dest": "#{BBOT_TOOLS}",
                "remote_src": True,
            },
        },
    ]

    common_web_ports = (
        80,
        443,
        # cloudflare HTTP
        8080,
        8880,
        2052,
        2082,
        2086,
        2095,
        # cloudflare HTTPS
        2053,
        2083,
        2087,
        2096,
        8443,
    )

    async def setup(self):
        self.skip_common_web = self.config.get("skip_common_web", True)
        return True

    async def filter_event(self, event):
        if self.skip_common_web:
            port_str = str(event.port)
            if event.port in self.common_web_ports or any(port_str.endswith(x) for x in ("080", "443")):
                return False, "port is a common web port and skip_common_web=True"
        return True

    async def handle_batch(self, *events):
        _input = {e.data: e for e in events}
        command = ["fingerprintx", "--json"]
        async for line in self.run_process_live(command, input=list(_input), stderr=subprocess.DEVNULL):
            try:
                j = json.loads(line)
            except Exception as e:
                self.debug(f'Error parsing line "{line}" as JSON: {e}')
                break
            ip = j.get("ip", "")
            host = j.get("host", ip)
            port = str(j.get("port", ""))
            protocol = j.get("protocol", "").upper()
            if not host and port and protocol:
                continue
            banner = j.get("metadata", {}).get("banner", "").strip()
            port_data = f"{host}:{port}"
            tags = set()
            if host and ip:
                tags.add(f"ip-{ip}")
            parent_event = _input.get(port_data)
            protocol_data = {"host": host, "protocol": protocol}
            if port:
                protocol_data["port"] = port
            if banner:
                protocol_data["banner"] = banner
            await self.emit_event(
                protocol_data,
                "PROTOCOL",
                parent=parent_event,
                tags=tags,
                context=f"{{module}} probed {port_data} and detected {{event.type}}: {protocol}",
            )



import asyncio
from OpenSSL import crypto
from contextlib import suppress

from bbot.errors import ValidationError
from bbot.modules.base import BaseModule
from bbot.core.helpers.async_helpers import NamedLock
from bbot.core.helpers.web.ssl_context import ssl_context_noverify


class sslcert(BaseModule):
    watched_events = ["OPEN_TCP_PORT"]
    produced_events = ["DNS_NAME", "EMAIL_ADDRESS"]
    flags = ["affiliates", "subdomain-enum", "email-enum", "active", "safe", "web-basic"]
    meta = {
        "description": "Visit open ports and retrieve SSL certificates",
        "created_date": "2022-03-30",
        "author": "@TheTechromancer",
    }
    options = {"timeout": 5.0, "skip_non_ssl": True}
    options_desc = {"timeout": "Socket connect timeout in seconds", "skip_non_ssl": "Don't try common non-SSL ports"}
    deps_apt = ["openssl"]
    deps_pip = ["pyOpenSSL~=25.3.0"]
    _module_threads = 25
    scope_distance_modifier = 1
    _priority = 2

    async def setup(self):
        self.timeout = self.config.get("timeout", 5.0)
        self.skip_non_ssl = self.config.get("skip_non_ssl", True)
        self.non_ssl_ports = (22, 53, 80)

        # sometimes we run into a server with A LOT of SANs
        # these are usually stupid and useless, so we abort based on a different threshold
        # depending on whether the parent event is in scope
        self.in_scope_abort_threshold = 50
        self.out_of_scope_abort_threshold = 10

        self.hosts_visited = set()
        self.ip_lock = NamedLock()
        return True

    async def filter_event(self, event):
        if self.skip_non_ssl and event.port in self.non_ssl_ports:
            return False, f"Port {event.port} doesn't typically use SSL"
        return True

    async def handle_event(self, event):
        _host = event.host
        if event.port:
            port = event.port
        else:
            port = 443

        # turn hostnames into IP address(es)
        if self.helpers.is_ip(_host):
            hosts = [_host]
        else:
            hosts = list(await self.helpers.resolve(_host))

        if event.scope_distance == 0:
            abort_threshold = self.in_scope_abort_threshold
        else:
            abort_threshold = self.out_of_scope_abort_threshold

        tasks = [self.visit_host(host, port) for host in hosts]
        async for task in self.helpers.as_completed(tasks):
            result = await task
            if not isinstance(result, tuple) or not len(result) == 3:
                continue
            dns_names, emails, (host, port) = result
            if len(dns_names) > abort_threshold:
                netloc = self.helpers.make_netloc(host, port)
                self.verbose(
                    f"Skipping Subject Alternate Names (SANs) on {netloc} because number of hostnames ({len(dns_names):,}) exceeds threshold ({abort_threshold})"
                )
                dns_names = dns_names[:1] + [n for n in dns_names[1:] if self.scan.in_scope(n)]
            for event_type, results in (("DNS_NAME", set(dns_names)), ("EMAIL_ADDRESS", emails)):
                for event_data in results:
                    if event_data is not None and event_data != event.data:
                        self.debug(f"Discovered new {event_type} via SSL certificate parsing: [{event_data}]")
                        try:
                            ssl_event = self.make_event(event_data, event_type, parent=event, raise_error=True)
                            parent_event = ssl_event.get_parent()
                            if parent_event.scope_distance == 0:
                                tags = ["affiliate"]
                            else:
                                tags = None
                            if ssl_event:
                                await self.emit_event(
                                    ssl_event,
                                    tags=tags,
                                    context=f"{{module}} parsed SSL certificate at {event.data} and found {{event.type}}: {{event.data}}",
                                )
                        except ValidationError as e:
                            self.hugeinfo(f'Malformed {event_type} "{event_data}" at {event.data}')
                            self.debug(f"Invalid data at {host}:{port}: {e}")

    def on_success_callback(self, event):
        parent_scope_distance = event.get_parent().scope_distance
        if parent_scope_distance == 0 and event.scope_distance > 0:
            event.add_tag("affiliate")

    async def visit_host(self, host, port):
        host = self.helpers.make_ip_type(host)
        netloc = self.helpers.make_netloc(host, port)
        host_hash = hash((host, port))
        dns_names = []
        emails = set()
        async with self.ip_lock.lock(host_hash):
            if host_hash in self.hosts_visited:
                self.debug(f"Already processed {host} on port {port}, skipping")
                return [], [], (host, port)
            else:
                self.hosts_visited.add(host_hash)

            host = str(host)

            # Connect to the host
            try:
                transport, _ = await asyncio.wait_for(
                    self.helpers.loop.create_connection(
                        lambda: asyncio.Protocol(), host, port, ssl=ssl_context_noverify
                    ),
                    timeout=self.timeout,
                )
            except asyncio.TimeoutError:
                self.debug(f"Timed out after {self.timeout} seconds while connecting to {netloc}")
                return [], [], (host, port)
            except Exception as e:
                log_fn = self.warning
                if isinstance(e, OSError):
                    log_fn = self.debug
                log_fn(f"Error connecting to {netloc}: {e}")
                return [], [], (host, port)
            finally:
                with suppress(Exception):
                    transport.close()

            # Get the SSL object
            try:
                ssl_object = transport.get_extra_info("ssl_object")
            except Exception as e:
                self.verbose(f"Error getting ssl_object: {e}", trace=True)
                return [], [], (host, port)

            # Get the certificate
            try:
                der = ssl_object.getpeercert(binary_form=True)
            except Exception as e:
                self.verbose(f"Error getting peer cert: {e}", trace=True)
                return [], [], (host, port)
            try:
                cert = crypto.load_certificate(crypto.FILETYPE_ASN1, der)
            except Exception as e:
                self.verbose(f"Error loading certificate: {e}", trace=True)
                return [], [], (host, port)
            issuer = cert.get_issuer()
            if issuer.emailAddress and self.helpers.regexes.email_regex.match(issuer.emailAddress):
                emails.add(issuer.emailAddress)
            subject = cert.get_subject()
            if subject.emailAddress and self.helpers.regexes.email_regex.match(subject.emailAddress):
                emails.add(subject.emailAddress)
            common_name = str(subject.commonName).lstrip("*.").lower()
            dns_names = set(self.get_cert_sans(cert))
            with suppress(KeyError):
                dns_names.remove(common_name)
            dns_names = [common_name] + list(dns_names)
        return dns_names, list(emails), (host, port)

    @staticmethod
    def get_cert_sans(cert):
        sans = []
        raw_sans = None
        ext_count = cert.get_extension_count()
        for i in range(0, ext_count):
            ext = cert.get_extension(i)
            short_name = str(ext.get_short_name())
            if "subjectAltName" in short_name:
                raw_sans = str(ext)
        if raw_sans is not None:
            for raw_san in raw_sans.split(","):
                hostname = raw_san.split(":", 1)[-1].strip().lower()
                # IPv6 addresses
                if hostname.startswith("[") and hostname.endswith("]"):
                    hostname = hostname.strip("[]")
                hostname = hostname.lstrip("*.")
                sans.append(hostname)
        return sans



from .base import ModuleTestBase


class TestGitlab_Com(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    modules_overrides = ["gitlab_com", "httpx", "social", "excavate"]

    async def setup_before_prep(self, module_test):
        module_test.httpserver.expect_request("/").respond_with_data("<a href='https://gitlab.org/veilidgroup'/>")
        module_test.httpx_mock.add_response(
            url="https://gitlab.org/api/v4/groups/veilidgroup/projects?simple=true",
            json=[
                {
                    "id": 55490429,
                    "description": None,
                    "name": "Veilid",
                    "name_with_namespace": "Veilid / Veilid",
                    "path": "veilid",
                    "path_with_namespace": "veilidgroup/veilid",
                    "created_at": "2024-03-03T05:22:53.169Z",
                    "default_branch": "master",
                    "tag_list": [],
                    "topics": [],
                    "ssh_url_to_repo": "git@gitlab.org:veilid/veilid.git",
                    "http_url_to_repo": "https://gitlab.org/veilidgroup/veilid.git",
                    "web_url": "https://gitlab.org/veilidgroup/veilid",
                    "readme_url": "https://gitlab.org/veilidgroup/veilid/-/blob/master/README.md",
                    "forks_count": 0,
                    "avatar_url": None,
                    "star_count": 0,
                    "last_activity_at": "2024-03-03T05:22:53.097Z",
                    "namespace": {
                        "id": 66882294,
                        "name": "veilidgroup",
                        "path": "veilidgroup",
                        "kind": "group",
                        "full_path": "veilidgroup",
                        "parent_id": None,
                        "avatar_url": "/uploads/-/system/group/avatar/66882294/signal-2023-07-04-192426_003.jpeg",
                        "web_url": "https://gitlab.org/groups/veilidgroup",
                    },
                },
            ],
        )

    def check(self, module_test, events):
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "SOCIAL"
                and e.data["platform"] == "gitlab"
                and e.data["profile_name"] == "veilidgroup"
                and e.data["url"] == "https://gitlab.org/veilidgroup"
            ]
        )
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "CODE_REPOSITORY"
                and "git" in e.tags
                and e.data["url"] == "https://gitlab.org/veilidgroup/veilid"
                and str(e.module) == "gitlab_com"
            ]
        )



from .base import ModuleTestBase


class TestPortfilter_disabled(ModuleTestBase):
    modules_overrides = []

    async def setup_before_prep(self, module_test):
        from bbot.modules.base import BaseModule

        class DummyModule(BaseModule):
            _name = "dummy_module"
            watched_events = ["DNS_NAME"]

            async def handle_event(self, event):
                if event.type == "DNS_NAME" and event.data == "blacklanternsecurity.com":
                    await self.emit_event(
                        "www.blacklanternsecurity.com:443",
                        "OPEN_TCP_PORT",
                        parent=event,
                        tags=["cdn-ip", "cdn-amazon"],
                    )
                    # when portfilter is enabled, this should be filtered out
                    await self.emit_event(
                        "www.blacklanternsecurity.com:8080",
                        "OPEN_TCP_PORT",
                        parent=event,
                        tags=["cdn-ip", "cdn-amazon"],
                    )
                    await self.emit_event("www.blacklanternsecurity.com:21", "OPEN_TCP_PORT", parent=event)

        module_test.scan.modules["dummy_module"] = DummyModule(module_test.scan)

    def check(self, module_test, events):
        open_ports = {event.data for event in events if event.type == "OPEN_TCP_PORT"}
        assert open_ports == {
            "www.blacklanternsecurity.com:443",
            "www.blacklanternsecurity.com:8080",
            "www.blacklanternsecurity.com:21",
        }


class TestPortfilter_enabled(TestPortfilter_disabled):
    modules_overrides = ["portfilter"]

    def check(self, module_test, events):
        # even though portfilter listens for URLs, enabling it should not automatically enable httpx
        assert "httpx" not in module_test.scan.modules
        open_ports = {event.data for event in events if event.type == "OPEN_TCP_PORT"}
        # we should be missing the 8080 port because it's a CDN and not in portfilter's allowed list of open ports
        assert open_ports == {"www.blacklanternsecurity.com:443", "www.blacklanternsecurity.com:21"}



from bbot.modules.templates.bucket import bucket_template


class bucket_google(bucket_template):
    """
    Adapted from https://github.com/RhinoSecurityLabs/GCPBucketBrute/blob/master/gcpbucketbrute.py
    """

    watched_events = ["DNS_NAME", "STORAGE_BUCKET"]
    produced_events = ["STORAGE_BUCKET", "FINDING"]
    flags = ["active", "safe", "cloud-enum", "web-basic"]
    meta = {
        "description": "Check for Google object storage related to target",
        "created_date": "2022-11-04",
        "author": "@TheTechromancer",
    }
    options = {"permutations": False}
    options_desc = {
        "permutations": "Whether to try permutations",
    }

    cloudcheck_provider_name = "Google"
    delimiters = ("", "-", ".", "_")
    base_domains = ["storage.googleapis.com"]
    bad_permissions = [
        "storage.buckets.get",
        "storage.buckets.list",
        "storage.buckets.create",
        "storage.buckets.delete",
        "storage.buckets.setIamPolicy",
        "storage.objects.get",
        "storage.objects.list",
        "storage.objects.create",
        "storage.objects.delete",
        "storage.objects.setIamPolicy",
    ]

    def filter_bucket(self, event):
        if not str(event.host).endswith(".googleapis.com"):
            return False, "bucket belongs to a different cloud provider"
        return True, ""

    def build_url(self, bucket_name, base_domain, region):
        return f"https://www.googleapis.com/storage/v1/b/{bucket_name}"

    async def check_bucket_open(self, bucket_name, url):
        bad_permissions = []
        try:
            list_permissions = "&".join(["=".join(("permissions", p)) for p in self.bad_permissions])
            url = f"https://www.googleapis.com/storage/v1/b/{bucket_name}/iam/testPermissions?" + list_permissions
            response = await self.helpers.request(url)
            permissions = response.json()
            if isinstance(permissions, dict):
                bad_permissions = list(permissions.get("permissions", {}))
        except Exception as e:
            self.info(f'Failed to enumerate permissions for bucket "{bucket_name}": {e}')
        msg = ""
        if bad_permissions:
            perms_str = ",".join(bad_permissions)
            msg = f"Open permissions on storage bucket ({perms_str})"
        return (msg, set())

    def check_bucket_exists(self, bucket_name, response):
        status_code = getattr(response, "status_code", 0)
        existent_bucket = status_code not in (0, 400, 404)
        return existent_bucket, set()



from .test_module_bucket_amazon import *


class TestBucket_Google(Bucket_Amazon_Base):
    provider = "google"
    random_bucket_1 = f"{random_bucket_name_1}.storage.googleapis.com"
    random_bucket_2 = f"{random_bucket_name_2}.storage.googleapis.com"
    random_bucket_3 = f"{random_bucket_name_3}.storage.googleapis.com"
    open_bucket_body = """{
  "kind": "storage#testIamPermissionsResponse",
  "permissions": [
    "storage.objects.create",
    "storage.objects.list"
  ]
}"""

    def bucket_setup(self):
        self.url_setup()
        self.website_body = f"""
        <a href="{self.url_1}"/>
        <a href="https://{self.random_bucket_2}"/>
        """

    def url_setup(self):
        self.url_1 = f"https://{random_bucket_name_1}.storage.googleapis.com"
        self.url_2 = f"https://www.googleapis.com/storage/v1/b/{random_bucket_name_2}/iam/testPermissions?&permissions=storage.buckets.get&permissions=storage.buckets.list&permissions=storage.buckets.create&permissions=storage.buckets.delete&permissions=storage.buckets.setIamPolicy&permissions=storage.objects.get&permissions=storage.objects.list&permissions=storage.objects.create&permissions=storage.objects.delete&permissions=storage.objects.setIamPolicy"
        self.url_3 = f"https://www.googleapis.com/storage/v1/b/{random_bucket_name_3}"



import pytest
import random
from bbot.core.helpers.misc import weighted_shuffle


class TestWeightedShuffleBenchmarks:
    """
    Benchmark tests for weighted_shuffle operations.

    This function is critical for BBOT's queue management, where it shuffles
    incoming queues based on module priority weights. Performance here directly
    impacts scan throughput and responsiveness.
    """

    def setup_method(self):
        """Setup common test data"""
        # Set deterministic seed for consistent benchmark results
        random.seed(42)  # Fixed seed for reproducible results

        # Generate test data of different sizes and complexity
        self.small_data = self._generate_small_dataset()
        self.medium_data = self._generate_medium_dataset()
        self.large_data = self._generate_large_dataset()
        self.priority_weights = self._generate_priority_weights()

    def _generate_small_dataset(self):
        """Generate small dataset (like few modules)"""
        return {"items": ["module_a", "module_b", "module_c"], "weights": [0.6, 0.3, 0.1]}

    def _generate_medium_dataset(self):
        """Generate medium dataset (like typical scan)"""
        items = [f"module_{i}" for i in range(20)]
        weights = [random.uniform(0.1, 1.0) for _ in range(20)]
        return {"items": items, "weights": weights}

    def _generate_large_dataset(self):
        """Generate large dataset (like complex scan with many modules)"""
        items = [f"module_{i}" for i in range(100)]
        weights = [random.uniform(0.1, 1.0) for _ in range(100)]
        return {"items": items, "weights": weights}

    def _generate_priority_weights(self):
        """Generate realistic priority weights (like BBOT module priorities)"""
        # BBOT uses priorities 1-5, where lower priority = higher weight
        # Weights are calculated as [5] + [6 - m.priority for m in modules]
        priorities = [5] + [6 - p for p in [1, 2, 3, 4, 5]] * 20  # 5 + 5*20 = 105 items
        items = [f"queue_{i}" for i in range(len(priorities))]
        return {"items": items, "weights": priorities}

    @pytest.mark.benchmark(group="weighted_shuffle")
    def test_typical_queue_shuffle(self, benchmark):
        """Benchmark weighted shuffle with typical BBOT scan workload"""

        def shuffle_typical():
            return weighted_shuffle(self.medium_data["items"], self.medium_data["weights"])

        result = benchmark(shuffle_typical)
        assert len(result) == 20
        assert all(item in result for item in self.medium_data["items"])

    @pytest.mark.benchmark(group="weighted_shuffle")
    def test_priority_queue_shuffle(self, benchmark):
        """Benchmark weighted shuffle with realistic BBOT priority weights"""

        def shuffle_priorities():
            return weighted_shuffle(self.priority_weights["items"], self.priority_weights["weights"])

        result = benchmark(shuffle_priorities)
        assert len(result) == len(self.priority_weights["items"])
        assert all(item in result for item in self.priority_weights["items"])



from .base import ModuleTestBase


class TestSecurityTxt(ModuleTestBase):
    targets = ["blacklanternsecurity.notreal"]
    modules_overrides = ["securitytxt", "speculate"]

    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://blacklanternsecurity.notreal/.well-known/security.txt",
            text="-----BEGIN PGP SIGNED MESSAGE-----\nHash: SHA512\n\nContact: mailto:joe.smith@blacklanternsecurity.notreal\nContact: mailto:vdp@example.com\nContact: https://vdp.example.com\nExpires: 2025-01-01T00:00:00.000Z\nPreferred-Languages: fr, en\nCanonical: https://blacklanternsecurity.notreal/.well-known/security.txt\nPolicy: https://example.com/cert\nHiring: https://www.careers.example.com\n-----BEGIN PGP SIGNATURE-----\n\nSIGNATURE\n\n-----END PGP SIGNATURE-----",
        )

    async def setup_after_prep(self, module_test):
        await module_test.mock_dns(
            {
                "blacklanternsecurity.notreal": {
                    "A": ["127.0.0.11"],
                },
            }
        )

    def check(self, module_test, events):
        assert any(e.type == "EMAIL_ADDRESS" and e.data == "joe.smith@blacklanternsecurity.notreal" for e in events), (
            "Failed to detect email address"
        )
        assert not any(
            e.type == "URL_UNVERIFIED" and e.data == "https://blacklanternsecurity.notreal/.well-known/security.txt"
            for e in events
        ), "Failed to filter Canonical URL to self"
        assert not any(str(e.data) == "vdp@example.com" for e in events)


class TestSecurityTxtEmailsFalse(TestSecurityTxt):
    config_overrides = {
        "scope": {"report_distance": 1},
        "modules": {"securitytxt": {"emails": False}},
    }

    def check(self, module_test, events):
        assert not any(e.type == "EMAIL_ADDRESS" for e in events), "Detected email address when emails=False"
        assert any(e.type == "URL_UNVERIFIED" and e.data == "https://vdp.example.com/" for e in events), (
            "Failed to detect URL"
        )
        assert any(e.type == "URL_UNVERIFIED" and e.data == "https://example.com/cert" for e in events), (
            "Failed to detect URL"
        )
        assert any(e.type == "URL_UNVERIFIED" and e.data == "https://www.careers.example.com/" for e in events), (
            "Failed to detect URL"
        )



import sys
import multiprocessing as mp

try:
    mp.set_start_method("spawn")
except Exception:
    start_method = mp.get_start_method()
    if start_method != "spawn":
        print(
            f"[WARN] Multiprocessing spawn method is set to {start_method}. This may negatively affect performance.",
            file=sys.stderr,
        )



from .base import ModuleTestBase


class TestEmails(ModuleTestBase):
    modules_overrides = ["emails", "emailformat", "skymem"]

    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://www.email-format.com/d/blacklanternsecurity.com/",
            text="""<a href="/cdn-cgi/l/email-protection" class="__cf_email__" data-cfemail="0a63646c654a68666b6961666b647e6f7864796f697f78637e7324696567">[email&#160;protected]</a>""",
        )
        module_test.httpx_mock.add_response(
            url="https://www.skymem.info/srch?q=blacklanternsecurity.com",
            text="<p>info@blacklanternsecurity.com</p>",
        )

    def check(self, module_test, events):
        assert 2 == len([e for e in events if e.data == "info@blacklanternsecurity.com"])
        email_file = module_test.scan.home / "emails.txt"
        emails = open(email_file).read().splitlines()
        # make sure deduping works as intended
        assert len(emails) == 1
        assert set(emails) == {"info@blacklanternsecurity.com"}



import json
from enum import IntEnum
from bbot.modules.base import BaseModule


class RetireJSSeverity(IntEnum):
    NONE = 0
    LOW = 1
    MEDIUM = 2
    HIGH = 3
    CRITICAL = 4

    @classmethod
    def from_string(cls, severity_str):
        try:
            return cls[severity_str.upper()]
        except (KeyError, AttributeError):
            return cls.NONE


class retirejs(BaseModule):
    watched_events = ["URL_UNVERIFIED"]
    produced_events = ["FINDING"]
    flags = ["active", "safe", "web-thorough"]
    meta = {
        "description": "Detect vulnerable/out-of-date JavaScript libraries",
        "created_date": "2025-08-19",
        "author": "@liquidsec",
    }
    options = {
        "version": "5.3.0",
        "node_version": "18.19.1",
        "severity": "medium",
    }
    options_desc = {
        "version": "retire.js version",
        "node_version": "Node.js version to install locally",
        "severity": "Minimum severity level to report (none, low, medium, high, critical)",
    }

    deps_ansible = [
        # Download Node.js binary (Linux x64)
        {
            "name": "Download Node.js binary (Linux x64)",
            "get_url": {
                "url": "https://nodejs.org/dist/v#{BBOT_MODULES_RETIREJS_NODE_VERSION}/node-v#{BBOT_MODULES_RETIREJS_NODE_VERSION}-linux-x64.tar.xz",
                "dest": "#{BBOT_TEMP}/node-v#{BBOT_MODULES_RETIREJS_NODE_VERSION}-linux-x64.tar.xz",
                "mode": "0644",
            },
        },
        # Extract Node.js binary (x64)
        {
            "name": "Extract Node.js binary (x64)",
            "unarchive": {
                "src": "#{BBOT_TEMP}/node-v#{BBOT_MODULES_RETIREJS_NODE_VERSION}-linux-x64.tar.xz",
                "dest": "#{BBOT_TOOLS}",
                "remote_src": True,
            },
        },
        # Remove existing node directory if it exists
        {
            "name": "Remove existing node directory",
            "file": {"path": "#{BBOT_TOOLS}/node", "state": "absent"},
        },
        # Rename extracted directory to 'node' (x64)
        {
            "name": "Rename Node.js directory (x64)",
            "command": "mv #{BBOT_TOOLS}/node-v#{BBOT_MODULES_RETIREJS_NODE_VERSION}-linux-x64 #{BBOT_TOOLS}/node",
        },
        # Set permissions on entire Node.js bin directory
        {
            "name": "Set permissions on Node.js bin directory",
            "file": {"path": "#{BBOT_TOOLS}/node/bin", "mode": "0755", "recurse": "yes"},
        },
        # Make Node.js binary executable
        {
            "name": "Make Node.js binary executable",
            "file": {"path": "#{BBOT_TOOLS}/node/bin/node", "mode": "0755"},
        },
        # Remove existing retirejs directory if it exists
        {
            "name": "Remove existing retirejs directory",
            "file": {"path": "#{BBOT_TOOLS}/retirejs", "state": "absent"},
        },
        # Create retire.js local directory
        {
            "name": "Create retire.js directory in BBOT_TOOLS",
            "file": {"path": "#{BBOT_TOOLS}/retirejs", "state": "directory", "mode": "0755"},
        },
        # Install retire.js locally using local Node.js
        {
            "name": "Install retire.js locally",
            "shell": "cd #{BBOT_TOOLS}/retirejs && #{BBOT_TOOLS}/node/bin/node #{BBOT_TOOLS}/node/lib/node_modules/npm/bin/npm-cli.js install --prefix . retire@#{BBOT_MODULES_RETIREJS_VERSION} --no-fund --no-audit --silent --no-optional",
            "args": {"creates": "#{BBOT_TOOLS}/retirejs/node_modules/.bin/retire"},
            "timeout": 600,
            "ignore_errors": False,
        },
        # Make retire script executable
        {
            "name": "Make retire script executable",
            "file": {"path": "#{BBOT_TOOLS}/retirejs/node_modules/.bin/retire", "mode": "0755"},
        },
        # Create retire cache directory
        {
            "name": "Create retire cache directory",
            "file": {"path": "#{BBOT_CACHE}/retire_cache", "state": "directory", "mode": "0755"},
        },
    ]

    accept_url_special = True
    scope_distance_modifier = 1
    _module_threads = 4

    async def setup(self):
        excavate_enabled = self.scan.config.get("excavate")
        if not excavate_enabled:
            return None, "retirejs will not function without excavate enabled"

        # Validate severity level
        valid_severities = ["none", "low", "medium", "high", "critical"]
        configured_severity = self.config.get("severity", "medium").lower()
        if configured_severity not in valid_severities:
            return (
                False,
                f"Invalid severity level '{configured_severity}'. Valid options are: {', '.join(valid_severities)}",
            )

        self.repofile = await self.helpers.download(
            "https://raw.githubusercontent.com/RetireJS/retire.js/master/repository/jsrepository-v4.json", cache_hrs=24
        )
        if not self.repofile:
            return False, "failed to download retire.js repository file"
        return True

    async def handle_event(self, event):
        js_file = await self.helpers.request(event.data)
        if js_file:
            js_file_body = js_file.text
            if js_file_body:
                js_file_body_saved = self.helpers.tempfile(js_file_body, pipe=False, extension="js")
                results = await self.execute_retirejs(js_file_body_saved)
                if not results:
                    self.warning("no output from retire.js")
                    return
                results_json = json.loads(results)
                if results_json.get("data"):
                    for file_result in results_json["data"]:
                        for component_result in file_result.get("results", []):
                            component = component_result.get("component", "unknown")
                            version = component_result.get("version", "unknown")
                            vulnerabilities = component_result.get("vulnerabilities", [])
                            for vuln in vulnerabilities:
                                severity = vuln.get("severity", "unknown")

                                # Filter by minimum severity level
                                min_severity = RetireJSSeverity.from_string(self.config.get("severity", "medium"))
                                vuln_severity = RetireJSSeverity.from_string(severity)
                                if vuln_severity < min_severity:
                                    self.debug(
                                        f"Skipping vulnerability with severity '{severity}' (below minimum '{min_severity.name.lower()}')"
                                    )
                                    continue

                                identifiers = vuln.get("identifiers", {})
                                summary = identifiers.get("summary", "Unknown vulnerability")
                                cves = identifiers.get("CVE", [])
                                description_parts = [
                                    f"Vulnerable JavaScript library detected: {component} v{version}",
                                    f"Severity: {severity.upper()}",
                                    f"Summary: {summary}",
                                    f"JavaScript URL: {event.data}",
                                ]
                                if cves:
                                    description_parts.append(f"CVE(s): {', '.join(cves)}")

                                below_version = vuln.get("below", "")
                                at_or_above = vuln.get("atOrAbove", "")
                                if at_or_above and below_version:
                                    description_parts.append(f"Affected versions: [{at_or_above} to {below_version})")
                                elif below_version:
                                    description_parts.append(f"Affected versions: [< {below_version}]")
                                elif at_or_above:
                                    description_parts.append(f"Affected versions: [>= {at_or_above}]")
                                description = " ".join(description_parts)
                                data = {
                                    "description": description,
                                    "severity": severity,
                                    "component": component,
                                    "url": event.parent.data["url"],
                                }
                                await self.emit_event(
                                    data,
                                    "FINDING",
                                    parent=event,
                                    context=f"{{module}} identified vulnerable JavaScript library {component} v{version} ({severity} severity)",
                                )

    async def filter_event(self, event):
        url_extension = getattr(event, "url_extension", "")
        if url_extension != "js":
            return False, f"it is a {url_extension} URL but retirejs only accepts js URLs"
        return True

    async def execute_retirejs(self, js_file):
        cache_dir = self.helpers.cache_dir / "retire_cache"
        retire_dir = self.scan.helpers.tools_dir / "retirejs"
        local_node_dir = self.scan.helpers.tools_dir / "node"

        # Use the retire binary directly with our local Node.js
        retire_binary_path = retire_dir / "node_modules" / ".bin" / "retire"
        command = [
            str(local_node_dir / "bin" / "node"),
            str(retire_binary_path),
            "--outputformat",
            "json",
            "--cachedir",
            str(cache_dir),
            "--path",
            js_file,
            "--jsrepo",
            str(self.repofile),
        ]

        proxy = self.scan.web_config.get("http_proxy")
        if proxy:
            command.extend(["--proxy", proxy])

        self.verbose(f"Running retire.js on {js_file}")
        self.verbose(f"retire.js command: {command}")

        result = await self.run_process(command)
        return result.stdout



import re

from bbot.modules.base import BaseModule


class git(BaseModule):
    watched_events = ["URL"]
    produced_events = ["FINDING", "CODE_REPOSITORY"]
    flags = ["active", "safe", "web-basic", "code-enum"]
    meta = {
        "description": "Check for exposed .git repositories",
        "created_date": "2023-05-30",
        "author": "@TheTechromancer",
    }

    in_scope_only = True

    fp_regex = re.compile(r"<html|<body", re.I)

    async def handle_event(self, event):
        base_url = event.data.rstrip("/")
        urls = {
            # look for git config in both
            self.helpers.urljoin(base_url, ".git/config"),
            self.helpers.urljoin(f"{base_url}/", ".git/config"),
        }
        async for url, response in self.helpers.request_batch(urls):
            text = getattr(response, "text", "")
            if not text:
                text = ""
            if text:
                if getattr(response, "status_code", 0) == 200 and "[core]" in text and not self.fp_regex.match(text):
                    description = f"Exposed .git config at {url}"
                    await self.emit_event(
                        {"host": str(event.host), "url": url, "description": description},
                        "FINDING",
                        event,
                        context="{module} detected {event.type}: {description}",
                    )
                    await self.emit_event(
                        {"url": url.rstrip("config")},
                        "CODE_REPOSITORY",
                        event,
                        tags="git_directory",
                        context="{module} detected {event.type}: {description}",
                    )



from bbot.modules.base import BaseModule


class GitLabBaseModule(BaseModule):
    """Common functionality for interacting with GitLab instances.

    This template is intended to be inherited by two concrete modules:
    1. ``gitlab_com``   – Handles public SaaS instances (gitlab.com / gitlab.org).
    2. ``gitlab_onprem`` – Handles self-hosted, on-premises GitLab servers.

    Both child modules share identical behaviour when talking to the GitLab
    REST API; they only differ in which events they are willing to accept.
    """

    # domains owned by GitLab
    saas_domains = ["gitlab.com", "gitlab.org"]

    async def setup(self):
        if self.options.get("api_key") is not None:
            await self.require_api_key()
        return True

    async def handle_social(self, event):
        """Enumerate projects belonging to a user or group profile."""
        username = event.data.get("profile_name", "")
        if not username:
            return
        base_url = self.get_base_url(event)
        urls = [
            # User-owned projects
            self.helpers.urljoin(base_url, f"api/v4/users/{username}/projects?simple=true"),
            # Group-owned projects
            self.helpers.urljoin(base_url, f"api/v4/groups/{username}/projects?simple=true"),
        ]
        for url in urls:
            await self.handle_projects_url(url, event)

    async def handle_projects_url(self, projects_url, event):
        for project in await self.gitlab_json_request(projects_url):
            project_url = project.get("web_url", "")
            if project_url:
                code_event = self.make_event({"url": project_url}, "CODE_REPOSITORY", tags="git", parent=event)
                await self.emit_event(
                    code_event,
                    context=f"{{module}} enumerated projects and found {{event.type}} at {project_url}",
                )
            namespace = project.get("namespace", {})
            if namespace:
                await self.handle_namespace(namespace, event)

    async def handle_groups_url(self, groups_url, event):
        for group in await self.gitlab_json_request(groups_url):
            await self.handle_namespace(group, event)

    async def gitlab_json_request(self, url):
        """Helper that performs an HTTP request and safely returns JSON list."""
        response = await self.api_request(url)
        if response is not None:
            try:
                json_data = response.json()
            except Exception:
                return []
            if json_data and isinstance(json_data, list):
                return json_data
        return []

    async def handle_namespace(self, namespace, event):
        namespace_name = namespace.get("path", "")
        namespace_url = namespace.get("web_url", "")
        namespace_path = namespace.get("full_path", "")

        if not (namespace_name and namespace_url and namespace_path):
            return

        namespace_url = self.helpers.parse_url(namespace_url)._replace(path=f"/{namespace_path}").geturl()

        social_event = self.make_event(
            {
                "platform": "gitlab",
                "profile_name": namespace_path,
                "url": namespace_url,
            },
            "SOCIAL",
            parent=event,
        )
        await self.emit_event(
            social_event,
            context=f'{{module}} found GitLab namespace ({{event.type}}) "{namespace_name}" at {namespace_url}',
        )

    # ------------------------------------------------------------------
    # Utility helpers
    # ------------------------------------------------------------------
    def get_base_url(self, event):
        base_url = event.data.get("url", "")
        if not base_url:
            base_url = f"https://{event.host}"
        return self.helpers.urlparse(base_url)._replace(path="/").geturl()



import csv
import ipaddress
from contextlib import suppress

from .csv import CSV
from bbot.core.helpers.misc import make_ip_type, is_ip, is_port, best_http_status

severity_map = {
    "INFO": 0,
    0: "N/A",
    1: "LOW",
    2: "MEDIUM",
    3: "HIGH",
    4: "CRITICAL",
    "N/A": 0,
    "LOW": 1,
    "MEDIUM": 2,
    "HIGH": 3,
    "CRITICAL": 4,
}


class asset_inventory(CSV):
    watched_events = [
        "OPEN_TCP_PORT",
        "DNS_NAME",
        "URL",
        "FINDING",
        "VULNERABILITY",
        "TECHNOLOGY",
        "IP_ADDRESS",
        "WAF",
        "HTTP_RESPONSE",
    ]
    produced_events = ["IP_ADDRESS", "OPEN_TCP_PORT"]
    meta = {
        "description": "Merge hosts, open ports, technologies, findings, etc. into a single asset inventory CSV",
        "created_date": "2022-09-30",
        "author": "@liquidsec",
    }
    options = {"output_file": "", "use_previous": False, "recheck": False, "summary_netmask": 16}
    options_desc = {
        "output_file": "Set a custom output file",
        "use_previous": "Emit previous asset inventory as new events (use in conjunction with -n <old_scan_name>)",
        "recheck": "When use_previous=True, don't retain past details like open ports or findings. Instead, allow them to be rediscovered by the new scan",
        "summary_netmask": "Subnet mask to use when summarizing IP addresses at end of scan",
    }

    header_row = [
        "Host",
        "Provider",
        "IP (External)",
        "IP (Internal)",
        "Open Ports",
        "HTTP Status",
        "HTTP Title",
        "Risk Rating",
        "Findings",
        "Technologies",
        "WAF",
        "DNS Records",
    ]
    filename = "asset-inventory.csv"

    async def setup(self):
        self.assets = {}
        self.use_previous = self.config.get("use_previous", False)
        self.recheck = self.config.get("recheck", False)
        self.summary_netmask = self.config.get("summary_netmask", 16)
        self.emitted_contents = False
        self._ran_hooks = False
        ret = await super().setup()
        return ret

    async def filter_event(self, event):
        if event._internal:
            return False, "event is internal"
        if event.type not in self.watched_events:
            return False, "event type is not in watched_events"
        if not self.scan.in_scope(event.host):
            return False, "event is not in scope"
        if "unresolved" in event.tags:
            return False, "event is unresolved"
        return True, ""

    async def handle_event(self, event):
        if (await self.filter_event(event))[0]:
            hostkey = _make_hostkey(event.host, event.resolved_hosts)
            if hostkey not in self.assets:
                self.assets[hostkey] = Asset(event.host, self.recheck)
            self.assets[hostkey].absorb_event(event)

    async def report(self):
        stats = {}
        totals = {}

        def increment_stat(stat, value):
            try:
                totals[stat] += 1
            except KeyError:
                totals[stat] = 1
            if stat not in stats:
                stats[stat] = {}
            try:
                stats[stat][value] += 1
            except KeyError:
                stats[stat][value] = 1

        def sort_key(asset):
            host = str(asset.host)
            is_digit = False
            with suppress(IndexError):
                is_digit = host[0].isdigit()
            return (is_digit, host)

        for asset in sorted(self.assets.values(), key=sort_key):
            findings_and_vulns = asset.findings.union(asset.vulnerabilities)
            ports = getattr(asset, "ports", set())
            ports = [str(p) for p in sorted([int(p) for p in asset.ports])]
            ips_all = getattr(asset, "ip_addresses", [])
            ips_external = sorted([str(ip) for ip in [i for i in ips_all if not i.is_private]])
            ips_internal = sorted([str(ip) for ip in [i for i in ips_all if i.is_private]])
            host = self.helpers.make_ip_type(getattr(asset, "host", ""))
            if host and isinstance(host, str):
                _, domain = self.helpers.split_domain(host)
                if domain:
                    increment_stat("Domains", domain)
            for ip in ips_all:
                net = ipaddress.ip_network(f"{ip}/{self.summary_netmask}", strict=False)
                increment_stat("IP Addresses", str(net))
            for port in ports:
                increment_stat("Open Ports", port)
            row = {
                "Host": host,
                "Provider": getattr(asset, "provider", ""),
                "IP (External)": ", ".join(ips_external),
                "IP (Internal)": ", ".join(ips_internal),
                "Open Ports": ", ".join(ports),
                "HTTP Status": asset.http_status_full,
                "HTTP Title": str(getattr(asset, "http_title", "")),
                "Risk Rating": severity_map[getattr(asset, "risk_rating", "")],
                "Findings": "\n".join(findings_and_vulns),
                "Technologies": "\n".join(str(x) for x in getattr(asset, "technologies", set())),
                "WAF": getattr(asset, "waf", ""),
                "DNS Records": ", ".join(sorted([str(r) for r in getattr(asset, "dns_records", [])])),
            }
            row.update(asset.custom_fields)
            self.writerow(row)

        for header in ("Domains", "IP Addresses", "Open Ports"):
            table_header = [header, ""]
            if header in stats:
                table = []
                stats_sorted = sorted(stats[header].items(), key=lambda x: x[-1], reverse=True)
                total = totals[header]
                for k, v in stats_sorted:
                    table.append([str(k), f"{v:,}/{total} ({v / total * 100:.1f}%)"])
                self.log_table(table, table_header, table_name=f"asset-inventory-{header}")

        if self._file is not None:
            self.info(f"Saved asset-inventory output to {self.output_file}")

    async def finish(self):
        if self.use_previous and not self.emitted_contents:
            self.emitted_contents = True
            if self.output_file.is_file():
                self.info(f"Emitting previous results from {self.output_file}")
                with open(self.output_file, newline="") as f:
                    c = csv.DictReader(f)
                    for row in c:
                        # yield to event loop to make sure we don't hold up the scan
                        await self.helpers.sleep(0)
                        host = row.get("Host", "").strip()
                        ips = row.get("IP (External)", "") + "," + row.get("IP (Internal)", "")
                        if not host or not ips:
                            continue
                        hostkey = _make_hostkey(host, ips)
                        asset = self.assets.get(hostkey, None)
                        if asset is None:
                            asset = Asset(host, self.recheck)
                            self.assets[hostkey] = asset
                        asset.absorb_csv_row(row)
                        self.add_custom_headers(list(asset.custom_fields))
                        if not is_ip(asset.host):
                            host_event = self.make_event(
                                asset.host, "DNS_NAME", parent=self.scan.root_event, raise_error=True
                            )
                            await self.emit_event(
                                host_event, context="{module} emitted previous result: {event.type}: {event.data}"
                            )
                            for port in asset.ports:
                                netloc = self.helpers.make_netloc(asset.host, port)
                                open_port_event = self.make_event(netloc, "OPEN_TCP_PORT", parent=host_event)
                                if open_port_event:
                                    await self.emit_event(
                                        open_port_event,
                                        context="{module} emitted previous result: {event.type}: {event.data}",
                                    )
                        else:
                            for ip in asset.ip_addresses:
                                ip_event = self.make_event(
                                    ip, "IP_ADDRESS", parent=self.scan.root_event, raise_error=True
                                )
                                await self.emit_event(
                                    ip_event, context="{module} emitted previous result: {event.type}: {event.data}"
                                )
                                for port in asset.ports:
                                    netloc = self.helpers.make_netloc(ip, port)
                                    open_port_event = self.make_event(netloc, "OPEN_TCP_PORT", parent=ip_event)
                                    if open_port_event:
                                        await self.emit_event(
                                            open_port_event,
                                            context="{module} emitted previous result: {event.type}: {event.data}",
                                        )
            else:
                self.warning(
                    f"use_previous=True was set but no previous asset inventory was found at {self.output_file}"
                )
        else:
            self._run_hooks()

    def _run_hooks(self):
        """
        modules can use self.asset_inventory_hook() to add custom functionality to asset_inventory
        the asset inventory module is passed in as the first argument to the method.
        """
        if not self._ran_hooks:
            self._ran_hooks = True
            for module in self.scan.modules.values():
                hook = getattr(module, "asset_inventory_hook", None)
                if hook is not None and callable(hook):
                    hook(self)


class Asset:
    def __init__(self, host, recheck):
        self.host = host
        self.ip_addresses = set()
        self.dns_records = set()
        self.ports = set()
        self.findings = set()
        self.vulnerabilities = set()
        self.status = "UNKNOWN"
        self.risk_rating = 0
        self.provider = ""
        self.waf = ""
        self.technologies = set()
        self.custom_fields = {}
        self.http_status = 0
        self.http_title = ""
        self.redirect_location = ""
        self.recheck = recheck

    def absorb_csv_row(self, row):
        # host
        host = make_ip_type(row.get("Host", "").strip())
        if host and not is_ip(host):
            self.host = host
        # ips
        self.ip_addresses = set(_make_ip_list(row.get("IP (External)", "")))
        self.ip_addresses.update(set(_make_ip_list(row.get("IP (Internal)", ""))))
        # If user requests a recheck dont import the following fields to force them to be rechecked
        if not self.recheck:
            # ports
            ports = [i.strip() for i in row.get("Open Ports", "").split(",")]
            self.ports.update({i for i in ports if i and is_port(i)})
            # findings
            findings = [i.strip() for i in row.get("Findings", "").splitlines()]
            self.findings.update({i for i in findings if i})
            # technologies
            technologies = [i.strip() for i in row.get("Technologies", "").splitlines()]
            self.technologies.update({i for i in technologies if i})
            # risk rating
            risk_rating = row.get("Risk Rating", "").strip()
            if risk_rating and risk_rating.isdigit() and int(risk_rating) > self.risk_rating:
                self.risk_rating = int(risk_rating)
            # provider
            provider = row.get("Provider", "").strip()
            if provider:
                self.provider = provider
        # custom fields
        for k, v in row.items():
            v = str(v)
            # update the custom field if it doesn't clash with our main fields
            # and if the new value isn't blank
            if v and k not in asset_inventory.header_row:
                self.custom_fields[k] = v

    def absorb_event(self, event):
        if not is_ip(event.host):
            self.host = event.host

        dns_children = getattr(event, "_dns_children", {})
        for rdtype, records in sorted(dns_children.items(), key=lambda x: x[0]):
            for record in sorted([str(r) for r in records]):
                self.dns_records.add(f"{rdtype}:{record}")

        http_status = getattr(event, "http_status", 0)
        update_http_status = bool(http_status) and best_http_status(http_status, self.http_status) == http_status
        if update_http_status:
            self.http_status = http_status
            if str(http_status).startswith("3"):
                if event.type == "HTTP_RESPONSE":
                    redirect_location = getattr(event, "redirect_location", "")
                    if redirect_location:
                        self.redirect_location = redirect_location
            else:
                self.redirect_location = ""

        if event.resolved_hosts:
            self.ip_addresses.update(set(_make_ip_list(event.resolved_hosts)))

        if event.port:
            self.ports.add(str(event.port))

        if event.type == "FINDING":
            location = event.data.get("url", event.data.get("host", ""))
            if location:
                self.findings.add(f"{location}:{event.data['description']}")

        if event.type == "VULNERABILITY":
            location = event.data.get("url", event.data.get("host", ""))
            if location:
                self.findings.add(f"{location}:{event.data['description']}:{event.data['severity']}")
                severity_int = severity_map.get(event.data.get("severity", "N/A"), 0)
                if severity_int > self.risk_rating:
                    self.risk_rating = severity_int

        if event.type == "TECHNOLOGY":
            self.technologies.add(event.data["technology"])

        if event.type == "WAF":
            if waf := event.data.get("waf", ""):
                if update_http_status or not self.waf:
                    self.waf = waf

        if event.type == "HTTP_RESPONSE":
            if title := event.data.get("title", ""):
                if update_http_status or not self.http_title:
                    self.http_title = title

        for tag in event.tags:
            if tag.startswith("cdn-") or tag.startswith("cloud-"):
                self.provider = tag
                break

    @property
    def hostkey(self):
        return _make_hostkey(self.host, self.ip_addresses)

    @property
    def http_status_full(self):
        return str(self.http_status) + (f" -> {self.redirect_location}" if self.redirect_location else "")


def _make_hostkey(host, ips):
    """
    We handle public and private IPs differently
    If the IPs are public, we dedupe by host
    If they're private, we dedupe by the IPs themselves
    """
    ips = _make_ip_list(ips)
    is_private = ips and all(is_ip(i) and i.is_private for i in ips)
    if is_private:
        return ",".join(sorted([str(i) for i in ips]))
    return str(host)


def _make_ip_list(ips):
    if isinstance(ips, str):
        ips = [i.strip() for i in ips.split(",")]
    ips = [make_ip_type(i) for i in ips if i and is_ip(i)]
    return ips



import os
import sys
import logging.handlers

loglevel_mapping = {
    "DEBUG": "DBUG",
    "TRACE": "TRCE",
    "VERBOSE": "VERB",
    "HUGEVERBOSE": "VERB",
    "INFO": "INFO",
    "HUGEINFO": "INFO",
    "SUCCESS": "SUCC",
    "HUGESUCCESS": "SUCC",
    "WARNING": "WARN",
    "HUGEWARNING": "WARN",
    "ERROR": "ERRR",
    "CRITICAL": "CRIT",
}
color_mapping = {
    "DEBUG": 242,  # grey
    "TRACE": 242,  # red
    "VERBOSE": 242,  # grey
    "INFO": 69,  # blue
    "HUGEINFO": 69,  # blue
    "SUCCESS": 118,  # green
    "HUGESUCCESS": 118,  # green
    "WARNING": 208,  # orange
    "HUGEWARNING": 208,  # orange
    "ERROR": 196,  # red
    "CRITICAL": 196,  # red
}
color_prefix = "\033[1;38;5;"
color_suffix = "\033[0m"


def colorize(s, level="INFO"):
    seq = color_mapping.get(level, 15)  # default white
    colored = f"{color_prefix}{seq}m{s}{color_suffix}"
    return colored


def log_to_stderr(msg, level="INFO", logname=True):
    """
    Print to stderr with BBOT logger colors
    """
    levelname = level.upper()
    if not any(x in sys.argv for x in ("-s", "--silent")):
        levelshort = f"[{loglevel_mapping.get(level, 'INFO')}]"
        levelshort = f"{colorize(levelshort, level=levelname)}"
        if levelname == "CRITICAL" or levelname.startswith("HUGE"):
            msg = colorize(msg, level=levelname)
        if logname:
            msg = f"{levelshort} {msg}"
        print(msg, file=sys.stderr)


class GzipRotatingFileHandler(logging.handlers.RotatingFileHandler):
    """
    A rotating file handler that compresses rotated files with gzip.
    Checks file size only periodically to improve performance.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._msg_count = 0
        self._check_interval = 1000  # Check size every 1000 messages

    def rotation_filename(self, default_name):
        """
        Modify the rotated filename to include .gz extension
        """
        return default_name + ".gz"

    def rotate(self, source, dest):
        """
        Compress the source file and move it to the destination.
        """
        import gzip

        with open(source, "rb") as f_in:
            with gzip.open(dest, "wb") as f_out:
                f_out.writelines(f_in)
        os.remove(source)

    def emit(self, record):
        """
        Emit a record, checking for rollover only periodically using modulo.
        """
        self._msg_count += 1

        # Only check for rollover periodically to save compute
        if self._msg_count % self._check_interval == 0:
            if self.shouldRollover(record):
                self.doRollover()

        # Continue with normal emit process
        super().emit(record)



import regex as re
from urllib.parse import urlparse
from bbot.modules.base import BaseModule


class ajaxpro(BaseModule):
    """
    Reference: https://mogwailabs.de/en/blog/2022/01/vulnerability-spotlight-rce-in-ajax.net-professional/
    """

    ajaxpro_regex = re.compile(r'<script.+src="([\/a-zA-Z0-9\._]+,[a-zA-Z0-9\._]+\.ashx)"')
    watched_events = ["HTTP_RESPONSE", "URL"]
    produced_events = ["VULNERABILITY", "FINDING"]
    flags = ["active", "safe", "web-thorough"]
    meta = {
        "description": "Check for potentially vulnerable Ajaxpro instances",
        "created_date": "2024-01-18",
        "author": "@liquidsec",
    }

    async def handle_event(self, event):
        if event.type == "URL" and "dir" in event.tags:
            await self.check_url_event(event)
        elif event.type == "HTTP_RESPONSE":
            await self.check_http_response_event(event)

    async def check_url_event(self, event):
        for stem in ["ajax", "ajaxpro"]:
            probe_url = f"{event.data}{stem}/whatever.ashx"
            probe = await self.helpers.request(probe_url)
            if probe and probe.status_code == 200:
                confirm_url = f"{event.data}a/whatever.ashx"
                confirm_probe = await self.helpers.request(confirm_url)
                if confirm_probe and confirm_probe.status_code != 200:
                    await self.emit_technology(event, probe_url)
                    await self.confirm_exploitability(probe_url, event)

    async def check_http_response_event(self, event):
        resp_body = event.data.get("body")
        if resp_body:
            match = await self.helpers.re.search(self.ajaxpro_regex, resp_body)
            if match:
                ajaxpro_path = match.group(0)
                await self.emit_technology(event, ajaxpro_path)
                await self.confirm_exploitability(ajaxpro_path, event)

    async def emit_technology(self, event, detection_url):
        url = event.data if event.type == "URL" else event.data["url"]
        await self.emit_event(
            {
                "host": str(event.host),
                "url": url,
                "technology": "ajaxpro",
            },
            "TECHNOLOGY",
            event,
            context=f"{self.meta['description']} discovered Ajaxpro instance ({event.type}) at {url} with trigger {detection_url}",
        )

    # Confirm exploitability of the detected Ajaxpro instance
    async def confirm_exploitability(self, detection_url, event):
        self.debug("Ajaxpro detected, attempting to confirm exploitability")
        parsed_url = urlparse(detection_url)
        base_url = f"{parsed_url.scheme}://{parsed_url.netloc}"
        path = parsed_url.path.rsplit("/", 1)[0]
        full_url = f"{base_url}{path}/AjaxPro.Services.ICartService,AjaxPro.2.ashx"

        # Payload and headers defined inline
        payload = {}
        headers = {"X-Ajaxpro-Method": "AddItem"}

        probe_response = await self.helpers.request(full_url, method="POST", headers=headers, json=payload)
        if probe_response:
            if "AjaxPro.Services.ICartService" and "MissingMethodException" in probe_response.text:
                await self.emit_event(
                    {
                        "host": str(event.host),
                        "severity": "CRITICAL",
                        "url": event.data if event.type == "URL" else event.data["url"],
                        "description": f"Ajaxpro Deserialization RCE (CVE-2021-23758) Trigger: [{full_url}]",
                    },
                    "VULNERABILITY",
                    event,
                    context=f"{self.meta['description']} discovered Ajaxpro instance ({event.type}) at {detection_url}",
                )



from bbot.modules.templates.censys import censys


class censys_ip(censys):
    """
    Query the Censys /v2/hosts/{ip} endpoint for associated hostnames, IPs, and URLs.
    """

    watched_events = ["IP_ADDRESS"]
    produced_events = [
        "IP_ADDRESS",
        "DNS_NAME",
        "URL_UNVERIFIED",
        "OPEN_TCP_PORT",
        "OPEN_UDP_PORT",
        "TECHNOLOGY",
        "PROTOCOL",
    ]
    flags = ["passive", "safe"]
    meta = {
        "description": "Query the Censys API for hosts by IP address",
        "created_date": "2026-01-26",
        "author": "@TheTechromancer",
        "auth_required": True,
    }
    options = {"api_key": "", "dns_names_limit": 100, "in_scope_only": True}
    options_desc = {
        "api_key": "Censys.io API Key in the format of 'key:secret'",
        "dns_names_limit": "Maximum number of DNS names to extract from dns.names (default 100)",
        "in_scope_only": "Only query in-scope IPs. If False, will query up to distance 1.",
    }
    scope_distance_modifier = 1

    async def setup(self):
        self.dns_names_limit = self.config.get("dns_names_limit", 100)
        self.warning(
            "This module may consume a lot of API queries. Unless you specifically want to query on each individual IP, we recommend using the censys_dns module instead."
        )
        return await super().setup()

    async def filter_event(self, event):
        in_scope_only = self.config.get("in_scope_only", True)
        max_scope_distance = 0 if in_scope_only else (self.scan.scope_search_distance + 1)
        if event.scope_distance > max_scope_distance:
            return False, "event is not in scope"
        return True

    async def handle_event(self, event):
        ip = str(event.host)
        url = f"{self.base_url}/v2/hosts/{ip}"

        resp = await self.api_request(url)
        if resp is None:
            self.debug(f"No response for {ip}")
            return

        if resp.status_code == 404:
            self.debug(f"No data found for {ip}")
            return

        if resp.status_code != 200:
            self.verbose(f"Non-200 status code ({resp.status_code}) for {ip}")
            return

        try:
            data = resp.json()
        except Exception as e:
            self.warning(f"Failed to parse JSON response for {ip}: {e}")
            return

        result = data.get("result", {})
        if not result:
            return

        # Track what we've already emitted to avoid duplicates
        seen = set()

        # Extract data from services
        for service in result.get("services", []):
            port = service.get("port")
            transport = service.get("transport_protocol", "TCP").upper()

            # Emit OPEN_TCP_PORT or OPEN_UDP_PORT for services with a port
            # QUIC uses UDP as transport, so treat it as UDP
            if port and (port, transport) not in seen:
                seen.add((port, transport))
                if transport in ("UDP", "QUIC"):
                    event_type = "OPEN_UDP_PORT"
                else:
                    event_type = "OPEN_TCP_PORT"
                await self.emit_event(
                    self.helpers.make_netloc(ip, port),
                    event_type,
                    parent=event,
                    context="{module} found open port on {event.parent.data}",
                )

            # Emit PROTOCOL for non-HTTP services
            # Use extended_service_name (more specific) falling back to service_name
            # Also check transport_protocol for protocols like QUIC
            service_name = service.get("extended_service_name") or service.get("service_name", "")
            # If service_name is UNKNOWN but transport_protocol is meaningful, use that
            if service_name.upper() == "UNKNOWN" and transport and transport not in ("TCP", "UDP"):
                service_name = transport
            if service_name and service_name.upper() not in ("HTTP", "HTTPS", "UNKNOWN"):
                protocol_key = ("protocol", service_name.upper(), port)
                if protocol_key not in seen:
                    seen.add(protocol_key)
                    protocol_data = {"host": str(event.host), "protocol": service_name}
                    if port:
                        protocol_data["port"] = port
                    await self.emit_event(
                        protocol_data,
                        "PROTOCOL",
                        parent=event,
                        context="{module} found {event.type}: {event.data[protocol]} on {event.parent.data}",
                    )

            # Extract URLs from HTTP services
            http_data = service.get("http", {})
            request = http_data.get("request", {})
            uri = request.get("uri")
            if uri and uri not in seen:
                seen.add(uri)
                await self.emit_event(
                    uri,
                    "URL_UNVERIFIED",
                    parent=event,
                    context="{module} found {event.data} in HTTP service of {event.parent.data}",
                )

            # Extract TLS certificate data
            tls_data = service.get("tls", {})
            certs = tls_data.get("certificates", {})
            leaf_data = certs.get("leaf_data", {})

            # Extract names from leaf_data.names
            for name in leaf_data.get("names", []):
                await self._emit_host(name, event, seen, "TLS certificate")

            # Extract common_name from leaf_data.subject
            subject = leaf_data.get("subject", {})
            for cn in subject.get("common_name", []):
                await self._emit_host(cn, event, seen, "TLS certificate subject")

            # Extract software/technologies
            for software in service.get("software", []):
                product = software.get("uniform_resource_identifier", software.get("product", ""))
                if product:
                    await self.emit_event(
                        {"technology": product, "host": str(event.host)},
                        "TECHNOLOGY",
                        parent=event,
                        context="{module} found {event.type}: {event.data[technology]} on {event.parent.data}",
                    )

        # Extract dns.names (limit to configured max)
        dns_data = result.get("dns", {})
        dns_names = dns_data.get("names", [])
        for name in dns_names[: self.dns_names_limit]:
            await self._emit_host(name, event, seen, "reverse DNS")

    async def _emit_host(self, host, event, seen, source):
        """Emit IP_ADDRESS or DNS_NAME for a host value."""
        # Validate and emit as DNS_NAME
        try:
            validated = self.helpers.validators.validate_host(host)
        except ValueError as e:
            self.debug(f"Error validating host {host} in {source}: {e}")
        if validated and validated not in seen:
            seen.add(validated)
            await self.emit_event(
                validated,
                "DNS_NAME",
                parent=event,
                context=f"{{module}} found {{event.data}} in {source} of {{event.parent.data}}",
            )



from .base import ModuleTestBase


class TestAzure_Tenant(ModuleTestBase):
    tenant_response = {
        "tenant_id": "cc74fc12-4142-400e-a653-f98bdeadbeef",
        "tenant_name": "blacklanternsecurity",
        "domain": "blacklanternsecurity.com",
        "email_domains": [
            "blacklanternsecurity.com",
            "blacklanternsecurity.onmicrosoft.com",
            "blsgvt.com",
            "o365.blacklanternsecurity.com",
        ],
    }

    async def setup_after_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://azmap.dev/api/tenant?domain=blacklanternsecurity.com&extract=true",
            json=self.tenant_response,
        )

    def check(self, module_test, events):
        assert any(
            e.type.startswith("DNS_NAME")
            and e.data == "blacklanternsecurity.onmicrosoft.com"
            and "affiliate" in e.tags
            for e in events
        )
        assert any(
            e.type == "AZURE_TENANT"
            and e.data["tenant-id"] == "cc74fc12-4142-400e-a653-f98bdeadbeef"
            and "blacklanternsecurity.onmicrosoft.com" in e.data["domains"]
            and "blacklanternsecurity" in e.data["tenant-names"]
            for e in events
        )
